// C Source File
// Created 10/11/2001; 5:56:20 PM

#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization
#define SAVE_SCREEN           // Save/Restore LCD Contents

#define USE_TI89
#define USE_TI92PLUS

#include <tigcclib.h>         // Include All Header Files

#include "CarSprites.h"

#define TYPE_EMPTY  0
#define TYPE_MYCARF 1
#define TYPE_MYCARB 2
#define TYPE_CAR1S1 3
#define TYPE_CAR1S2 4
#define TYPE_CAR2S1 5
#define TYPE_CAR2S2 6
#define TYPE_CAR2S3 7
#define TYPE_CAR3S1 8
#define TYPE_CAR3S2 9
#define TYPE_CAR3S3 10
#define TYPE_CAR3S4 11

#define TF_N 0x10
#define TF_E 0x20
#define TF_S 0x40
#define TF_W 0x80

#define TF_0   0
#define TF_90  1
#define TF_180 2
#define TF_270 3

#define TF_EXIT 0x08

#define TILE_MAKE(t, Tile){

typedef struct Tile			// My Tile information
{
  unsigned char ucType;		// Defines the type of car. 0 for none, 1 for my car front, 2 for my car back, etc.
  unsigned char ucFlags;  	// Flags. Upper bits holds connection info, lower, orientation.
} Tile;

typedef struct TileSet
{
	unsigned char width, height;
	unsigned char exitDist;
	char exitX, exitY;
	unsigned char myCarCount;
	struct Tile **Tiles;
} TileSet;

#define INFO_TO_WIDTH(i)  	((i >> 4) & 0x0F)
#define INFO_TO_HEIGHT(i)	(i & 0x0F)
#define INFO_TO_EXIT(i)		(i >> 8)
#define INFO_MAKE(x, w, h)	(((x) << 8) | ((w) << 4) | (h))

#define TYPE_MYCAR	0
#define TYPE_CAR1	1
#define TYPE_CAR2	2
#define TYPE_CAR3	3
#define TYPE_LIMO	2
#define TYPE_SEMI	3

#define ROT_N	0
#define ROT_E	1
#define ROT_S	2
#define ROT_W	3

#define CAR_TO_X(car)		((car >> 4) & 0x0F)
#define CAR_TO_Y(car)		(car & 0x0F)
#define CAR_TO_TYPE(car)	((car >> 8) & 0x0F)
#define CAR_TO_ROT(car)		(car >> 12)
#define CAR_MAKE(r, t, x, y) (((r) << 12) | ((t) << 8) | ((x) << 4) | (y));


// Cars have the form:
//  unsigned int  car;       %00rrttttxxxxyyyy r=rotation, t=type, x=x coord, y=y coord
typedef unsigned short Car;

typedef struct Level
{
  unsigned short info;  // %00xxxxxx wwwwhhhh
  						//	 x = Distance from top left on perimeter to exit. POSITIVE ONLY
  						//	 w = Width of parking lot, h = Height of parking lot. 92+: Max Height: 7; Max Width: 14. 89: Max Height: 6; Max Width: 9.
  unsigned char  nCars; // Number of cars
  unsigned short *cars;	// Pointer to cars.
} Level;

typedef struct LevelSet
{
	unsigned char nLevels;
	Level *levels;
} LevelSet;

LevelSet ls = {0, NULL};
TileSet ts = {0, 0, 0, 0, 0, 0, NULL};
SYM_ENTRY* *LevelSets = NULL;
unsigned char nLevelSets = 0;
void *kbq = NULL;
unsigned short dx = 0;
unsigned short dy = 0;
unsigned char* lPlane = NULL;
unsigned char* dPlane = NULL;
unsigned char MAXWIDTH = 0;
unsigned char MAXHEIGHT = 0;

void RSprite16(short x, short y, const unsigned short *sprite, void *vm_addr, short mode, short angle);

unsigned char LoadTS(unsigned char l);
void UnloadTS();
static inline void DisplayTS();
void PlayTS();
unsigned char MakeFlags(unsigned char segc, unsigned char segi, unsigned char rot);

static inline void FindLS();
void LoadLS(unsigned char);
void UnloadLS();

void DisplayTile(unsigned char, unsigned char);
void InvertTile(unsigned char, unsigned char);
void DeleteTile(unsigned char, unsigned char);

void MoveH(unsigned char, unsigned char, char, unsigned char);
void MoveV(unsigned char, unsigned char, char, unsigned char);
#define MoveLeft(x, y)  MoveH(x, y, -1, TF_E)
#define MoveRight(x, y) MoveH(x, y, 1, TF_W)
#define MoveUp(x, y)    MoveV(x, y, -1, TF_S)
#define MoveDown(x, y)  MoveV(x, y, 1, TF_N)

char Menu(char);
void DrawMenu(unsigned char, char);

// Main Function
void _main(void)
{
  char choice;
  kbq = kbd_queue();

  
#ifndef TI92_ONLY
  if(TI89)
  {
    MAXWIDTH = 9; MAXHEIGHT = 6;
  } else {
    MAXWIDTH = 14; MAXHEIGHT = 7;
  }
#endif
  
  LevelSets = (SYM_ENTRY**)malloc(0);
  FindLS();
  if(nLevelSets == 0)
  {
    free(LevelSets);
	  exit(850);
	}
  choice = Menu(0);
  while (choice >= 0)
  {
    GrayOn();
  lPlane = (unsigned char*)GetPlane(LIGHT_PLANE);
  dPlane = (unsigned char*)GetPlane(DARK_PLANE);
#undef SetPlane
#define SetPlane(x) PortSet(x, 239, 127)
#undef LIGHT_PLANE
#undef DARK_PLANE
#define LIGHT_PLANE lPlane
#define DARK_PLANE dPlane

    if(LoadTS(choice)) {
      DisplayTS();
      PlayTS();
      UnloadTS();
    }
#ifndef TI92_ONLY
    else
    {
      #define levWarningMessage "LEVEL TOO LARGE FOR TI89!"
      ClearScreen();
      FontSetSys(F_4x6);
      SetPlane(DARK_PLANE);
      DrawStr((LCD_WIDTH - DrawStrWidth(levWarningMessage, F_4x6)) >> 1, (LCD_HEIGHT - 4) >> 1, levWarningMessage, A_NORMAL);
      SetPlane(LIGHT_PLANE);
      DrawStr((LCD_WIDTH - DrawStrWidth(levWarningMessage, F_4x6)) >> 1, (LCD_HEIGHT - 4) >> 1, levWarningMessage, A_NORMAL);
      ngetchx();
    }
#endif
	  GrayOff();

    choice = Menu(choice);
  }
  UnloadLS();
  free(LevelSets);
  
}

char Menu(char choice)
{
  short key = 0;
  static char curLvlSet = -1;
  if(curLvlSet == -1)
  {
    LoadLS(0);
    curLvlSet = 0;
  }
  choice++;
  while(key != KEY_ESC)
  {
    DrawMenu(choice, curLvlSet);
    while(OSdequeue((unsigned short*)&key, kbq));
    if(key == KEY_UP)
    {
      if(++choice >= ls.nLevels + 1) choice = 1;
    } else if(key == KEY_DOWN) {
      if(--choice == 0) choice = ls.nLevels;
    } else if(key == KEY_RIGHT) {
      if(++curLvlSet == nLevelSets)
        curLvlSet = 0;
      UnloadLS();
      LoadLS(curLvlSet);
      choice = 1;
    } else if(key == KEY_LEFT) {
      if(--curLvlSet < 0)
				curLvlSet = nLevelSets - 1;      
      UnloadLS();
      LoadLS(curLvlSet);
      choice = 1;
    } else if(key == KEY_ENTER)
    {
      return choice - 1;
    } else if(key == KEY_F1)
    {
    	ClearScreen();
    	DrawStr(2, 22, "Remove the dark car from the parking lot.", A_NORMAL);
    	DrawStr(2, 30, "Move the cursor to the end of a car and press", A_NORMAL);
    	DrawStr(2, 38, "F1 or Enter to move it.  Rearrange the cars", A_NORMAL);
    	DrawStr(2, 46, "until you're out!", A_NORMAL);
    	while(OSdequeue((unsigned short*)&key, kbq));
    	key = 0;
    }
	}

  return -1;
}      
  
void DrawMenu(unsigned char choice, char curLvlSet)
{
  char lev[11];
  char levSet[11];
  unsigned char offset = 0;
  levSet[0] = 0x11;
  for(;offset < 8 && LevelSets[curLvlSet]->name[offset] != 0; offset++)
    levSet[offset + 1] = LevelSets[curLvlSet]->name[offset];
  levSet[++offset] = 0x12;
  levSet[++offset] = 0;
  #define title "Car Jam"
  #define author "Cody Cutrer (programmer066)"
  #define helpstring "F1 - Help"
  sprintf(lev, "Level: %c%02d%c", 0x14, choice, 0x13);
  ClearScreen();
  FontSetSys(F_8x10);
//  DrawStr((LCD_WIDTH - DrawStrWidth(title, F_8x10)) >> 1, 5, title, A_NORMAL);
  DrawStr((LCD_WIDTH - 56) >> 1, 5, title, A_NORMAL);
  FontSetSys(F_6x8);
//  DrawStr((LCD_WIDTH - DrawStrWidth(lev, F_6x8)) >> 1, 30, lev, A_NORMAL);
  DrawStr((LCD_WIDTH - 60) >> 1, 38, lev, A_NORMAL);
  DrawStr((LCD_WIDTH - DrawStrWidth(levSet, F_6x8)) >> 1, 28, levSet, A_NORMAL);
  FontSetSys(F_4x6);
  DrawStr((LCD_WIDTH - DrawStrWidth(author, F_4x6)) >> 1, LCD_HEIGHT - 10, author, A_NORMAL);
  DrawStr((LCD_WIDTH - DrawStrWidth(helpstring, F_4x6)) >> 1, LCD_HEIGHT >> 1, helpstring, A_NORMAL);
}

static inline unsigned char IsLS(SYM_ENTRY* sym)
{
  unsigned char* f = (unsigned char*)HLock(sym->handle);
  unsigned short typeoffset = *f++ << 8; // Read in file size
  unsigned char isls = 0;
  typeoffset |= *f++;
  f += typeoffset - 2 - 3; // seek to end of file (subtract 2 because we just read in 2 for file size) - 3 to get to type
  if(!strcmp(f, "car")) isls = 1;
  HeapUnlock(sym->handle);
  return isls;
}  

void FindLS()
{
  SYM_ENTRY *sym = SymFindFirst(NULL, 2);
  nLevelSets = 0;
  while(sym)
  {
    if(IsLS(sym))  // If the current symbol is a levelset
    {
      LevelSets = (SYM_ENTRY**)realloc(LevelSets, sizeof(SYM_ENTRY*) * ++nLevelSets); // Add room in our LS array
      LevelSets[nLevelSets - 1] = sym; // Save it
    }
    sym = SymFindNext();  // Go to the next one
  }
}  

void LoadLS(unsigned char whichls)
{
//	SYM_ENTRY *sym = SymFindPtr($(carlvl), 0);
	unsigned char i = 0;
	unsigned char* f = (unsigned char*)HLock(LevelSets[whichls]->handle) + 2;
	ls.nLevels = *f++;
  ls.levels = (Level*)malloc(sizeof(Level) * ls.nLevels);


	for(; i < ls.nLevels; i++)
	{
	  ls.levels[i].info = *f++ << 8;
	  ls.levels[i].info |= *f++;
		ls.levels[i].nCars = *f++;
  	ls.levels[i].cars = (Car*)malloc(ls.levels[i].nCars << 1);
  	memcpy(ls.levels[i].cars, f, ls.levels[i].nCars << 1);
    f += ls.levels[i].nCars << 1;
	}
	HeapUnlock(LevelSets[whichls]->handle);
}

void UnloadLS()
{
  unsigned char i=0;
  for(; i < ls.nLevels; i++)
  {
    free(ls.levels[i].cars);
  }
  free(ls.levels);
}

void UnloadTS()
{
  unsigned char i = 0;
  for(; i < ts.width; i++)
  {
  	free(ts.Tiles[i]);
  }
  free(ts.Tiles);
}

unsigned char LoadTS(unsigned char l)
{
  unsigned char i = 0;
  
  ts.width = INFO_TO_WIDTH(ls.levels[l].info);
  ts.height = INFO_TO_HEIGHT(ls.levels[l].info);
#ifndef TI92_ONLY
  if(ts.width > MAXWIDTH || ts.height > MAXHEIGHT)
    return 0;
#endif
  ts.exitDist = INFO_TO_EXIT(ls.levels[l].info);
  if(ts.exitDist < ts.width) // exit on top
  {
    ts.exitX = ts.exitDist; ts.exitY = -1;
  } else if(ts.exitDist < ts.width + ts.height)
  {
    ts.exitX = ts.width; ts.exitY = ts.exitDist - ts.width;
  } else if(ts.exitDist < (ts.width << 1) + ts.height) {
    ts.exitX = -ts.exitDist + (ts.width << 1) + ts.height - 1; ts.exitY = ts.height;
  } else
  {
    ts.exitX = -1; ts.exitY = -ts.exitDist + (ts.width << 1) + (ts.height << 1) - 1;
  }
  ts.myCarCount = 0;
  ts.Tiles = (Tile**)malloc(sizeof(Tile*) * ts.width);

  for(; i < ts.width; i++)
  {
    ts.Tiles[i] = (Tile*)malloc(sizeof(Tile) * ts.height);
    memset(ts.Tiles[i], 0, sizeof(Tile) * ts.height);
  }

  for(i = 0; i < ls.levels[l].nCars; i++)
  {
    Tile t = {0, 0};
    Car c = ls.levels[l].cars[i];
    #define type CAR_TO_TYPE(c)
    unsigned char segc = (type == TYPE_CAR2) ? 3 : ((type == TYPE_CAR3) ? 4 : 2); // Segment count
    unsigned char segi = 0;	// Segment index
    unsigned char typeoff = (type == TYPE_CAR2) ? TYPE_CAR2S1 : ((type == TYPE_CAR3) ? TYPE_CAR3S1 : ((type == TYPE_CAR1) ? TYPE_CAR1S1 : TYPE_MYCARF)); // Segment Type Offset
    #define reverse (CAR_TO_ROT(c) == ROT_S || CAR_TO_ROT(c) == ROT_E) // If the car's front is not N or W
    #define horiz (CAR_TO_ROT(c) == ROT_E || CAR_TO_ROT(c) == ROT_W) // If the car lies horizontally
    if(type == TYPE_MYCAR)
      ts.myCarCount += 2;
    for(; segi < segc; segi++)
    {
      if(reverse) {
        t.ucType = segc - segi + typeoff - 1;
      } else {
      	t.ucType = segi + typeoff;
      }
      t.ucFlags = MakeFlags(segc, segi, CAR_TO_ROT(c));
      if(horiz != 0) {
        ts.Tiles[CAR_TO_X(c) + segi][CAR_TO_Y(c)] = t;
      } else {
        ts.Tiles[CAR_TO_X(c)][CAR_TO_Y(c) + segi] = t;
      }
    }
  }
  
  return 1;
}

unsigned char MakeFlags(unsigned char segc, unsigned char segi, unsigned char rot)
{
  unsigned char flags;

  if(rot == ROT_N || rot == ROT_S)
	{
    flags = TF_N | TF_S;
	  if(segi == 0)
	    flags &= ~TF_N;
	  else if(segi == segc - 1)
	    flags &= ~TF_S;
	  
	  if(rot == ROT_N)
	    flags |= TF_270;
	  else
	    flags |= TF_90;
  } else {
	  flags = TF_E | TF_W;
 	  if(segi == 0)
	    flags &= ~TF_W;
	  else if(segi == segc - 1)
	    flags &= ~TF_E;
	    
	  if(rot == ROT_W)
	    flags |= TF_180;
  }
 
  return flags; 
} 	
#define FULLSCREEN  &(SCR_RECT){{0, 0, 239, 127}}
void DisplayTS()
{
	unsigned char i, j;
	unsigned char dxw, dyw;
	WIN_RECT border;

	dx = ((LCD_WIDTH - (ts.width << 4)) >> 1);		// x offset. The bit shifts just multiply by 16
	dy = (LCD_HEIGHT - (ts.height << 4)) >> 1;	// y offset
	dxw = dx + (ts.width << 4);									// x offset plus width	
	dyw = dy + (ts.height << 4);								// y offset plus height
	border.x0 = dx - 2; border.y0 = dy - 2; border.x1 = dxw + 1; border.y1 = dyw + 1;	// Calculates a bounding rectangle for the tileset
	
	ClearScreen();
	
	for(i = 0; i < ts.height; i++)
		for(j = 0; j < ts.width; j++)
			DisplayTile(j, i);
			
	SetPlane(LIGHT_PLANE);
	DrawClipRect(&border, FULLSCREEN, A_NORMAL | B_NORMAL);
	SetPlane(DARK_PLANE);
	DrawClipRect(&border, FULLSCREEN, A_NORMAL | B_NORMAL);
	if(ts.exitDist < ts.width)	// If the exit is on the top
	{
		DrawLine(dx + (ts.exitX << 4), dy - 2, dx + (ts.exitX << 4) + 16, dy - 2, A_REVERSE); // Erase the border
	} else if(ts.exitDist < ts.width + ts.height)
	{
		#define offset  (ts.exitY << 4)
		DrawLine(dxw + 1, dy + offset, dxw + 1, dy + offset + 16, A_REVERSE);
		#undef offset
	} else if(ts.exitDist < ((ts.width << 1) + ts.height)) 	// else if exit is on the bottom
	{
		#define offset (ts.exitX << 4)	// Calculate an x offset
		DrawLine(dx + offset, dyw + 1, dx + offset + 16, dyw + 1, A_REVERSE);
		#undef offset
	} else if(ts.exitDist < ((ts.width + ts.height) << 1))	// else if exit is on the left
	{
		#define offset (ts.exitY << 4)
		DrawLine(dx - 2, dy + offset, dx - 2, dy + offset + 16, A_REVERSE);
		#undef offset
	}	
}
	
void PlayTS()
{
	char sx = 0, sy = 0; // Selected coordinates
	short key = 0;

	InvertTile(sx, sy);	
	while(key != KEY_ESC)
	{
		while(OSdequeue((unsigned short*)&key, kbq));
//		if(key >= 
		if(key == KEY_RIGHT)
		{
			InvertTile(sx, sy);
			if(++sx >= ts.width) sx = 0;
			InvertTile(sx, sy);
		} else if(key == KEY_LEFT)
		{
			InvertTile(sx, sy);
			if(--sx < 0) sx = ts.width - 1;
			InvertTile(sx, sy);
		} else if(key == KEY_DOWN)
		{
			InvertTile(sx, sy);
			if(++sy >= ts.height) sy = 0;
			InvertTile(sx, sy);
		} else if(key == KEY_UP)
		{
			InvertTile(sx, sy);
			if(--sy < 0) sy = ts.height - 1;
			InvertTile(sx, sy);
		} else if(key == KEY_ENTER || key == KEY_F1)
		{
			if((sx > 0) && (ts.Tiles[sx-1][sy].ucType == TYPE_EMPTY) && (ts.Tiles[sx][sy].ucFlags & TF_E))
			{
				MoveLeft(sx, sy);
				OSenqueue(KEY_LEFT, kbq);
			}
			else if((sx < ts.width - 1) && (ts.Tiles[sx+1][sy].ucType == TYPE_EMPTY) && (ts.Tiles[sx][sy].ucFlags & TF_W))
			{
				MoveRight(sx, sy);
				OSenqueue(KEY_RIGHT, kbq);
			}
			else if((sy > 0) && (ts.Tiles[sx][sy-1].ucType == TYPE_EMPTY) && (ts.Tiles[sx][sy].ucFlags & TF_S))
			{
				MoveUp(sx, sy);
				OSenqueue(KEY_UP, kbq);
			}
			else if((sy < ts.height - 1) && (ts.Tiles[sx][sy+1].ucType == TYPE_EMPTY) && (ts.Tiles[sx][sy].ucFlags & TF_N))
			{
				MoveDown(sx, sy);
				OSenqueue(KEY_DOWN, kbq);
			}
			
			else if((sx == 0) && (ts.exitX == sx - 1) && (ts.exitY == sy) && (ts.Tiles[sx][sy].ucFlags & TF_E)) // Exit on left
				DeleteTile(sx, sy);
				
			else if((sx == ts.width - 1) && (ts.exitX == sx + 1) && (ts.exitY == sy) && (ts.Tiles[sx][sy].ucFlags & TF_W)) // Exit on right
				DeleteTile(sx, sy);
				
			else if((sy == 0) && (ts.exitY == sy - 1) && (ts.exitX == sx) && (ts.Tiles[sx][sy].ucFlags & TF_S)) // Exit on top
				DeleteTile(sx, sy);
				
			else if((sy == ts.height - 1) && (ts.exitY == sy + 1) && (ts.exitX == sx) && (ts.Tiles[sx][sy].ucFlags & TF_N)) // Exit on bottom
				DeleteTile(sx, sy);


			if(ts.myCarCount <= 0)
			{
				#define msg "You Win!!"
				FontSetSys(F_8x10);
				InvertTile(sx, sy);
				ClearScreen();
				DrawStr((LCD_WIDTH - DrawStrWidth(msg, F_8x10)) >> 1, (LCD_HEIGHT - 10) >> 1, msg, A_NORMAL);
				SetPlane(LIGHT_PLANE);
				DrawStr((LCD_WIDTH - DrawStrWidth(msg, F_8x10)) >> 1, (LCD_HEIGHT - 10) >> 1, msg, A_NORMAL);
			  ngetchx();
				return;
			}
		}
	}
}

void DisplayTile(unsigned char sx, unsigned char sy)
{
//  if(ts.Tiles[sx][sy].ucType == 0) return;
	RSprite16((sx	<< 4) + dx, (sy << 4) + dy, Sprites_C0[ts.Tiles[sx][sy].ucType], lPlane, SPRT_XOR, (ts.Tiles[sx][sy].ucFlags & 0x0F) * 90);
	RSprite16((sx	<< 4) + dx, (sy << 4) + dy, Sprites_C1[ts.Tiles[sx][sy].ucType], dPlane, SPRT_XOR, (ts.Tiles[sx][sy].ucFlags & 0x0F) * 90);		
}

// FUNCTION MODIFIED FROM InvertGrayRect2b from ExtGraph v0.86 by Thomas Nussbaumer
void InvertTile(unsigned char sx,unsigned char sy) {
    SCR_RECT sr = {{dx + (sx << 4),dy + (sy << 4),dx + (sx<<4)+15,dy+(sy<<4)+15}};

    PortSet(lPlane,239,127);
    ScrRectFill(&sr,FULLSCREEN,A_XOR);
    PortSet(dPlane,239,127);
    ScrRectFill(&sr,FULLSCREEN,A_XOR);
}

void MoveH(unsigned char sx, unsigned char sy, char dsx, unsigned char tf)
{
	ts.Tiles[sx + dsx][sy] = ts.Tiles[sx][sy];
	DisplayTile(sx, sy);
	DisplayTile(sx + dsx, sy);
	ts.Tiles[sx][sy].ucType = 0;
	ts.Tiles[sx][sy].ucFlags = 0;
	if(ts.Tiles[sx + dsx][sy].ucFlags & tf)
		MoveH(sx - dsx, sy, dsx, tf);
}

void MoveV(unsigned char sx, unsigned char sy, char dsy, unsigned char tf)
{
	ts.Tiles[sx][sy+dsy] = ts.Tiles[sx][sy];
	DisplayTile(sx, sy);
	DisplayTile(sx, sy+dsy);
	ts.Tiles[sx][sy].ucType = 0;
	ts.Tiles[sx][sy].ucFlags = 0;
	if(ts.Tiles[sx][sy+dsy].ucFlags & tf)
		MoveV(sx, sy-dsy,dsy,tf);
}

void DeleteTile(unsigned char sx, unsigned char sy)
{
	unsigned char ucFlags = ts.Tiles[sx][sy].ucFlags;

	DisplayTile(sx, sy);
	if(ts.Tiles[sx][sy].ucType == TYPE_MYCARF || ts.Tiles[sx][sy].ucType == TYPE_MYCARB)
		ts.myCarCount--;										// If it was my car, decrement that number for tileset
	ts.Tiles[sx][sy].ucType = TYPE_EMPTY;
	ts.Tiles[sx][sy].ucFlags = 0;
	if(ucFlags & TF_N)
		DeleteTile(sx, sy-1);
	if(ucFlags & TF_E)
		DeleteTile(sx+1, sy);
	if(ucFlags & TF_S)
		DeleteTile(sx, sy+1);
	if(ucFlags & TF_W)
		DeleteTile(sx-1, sy);
}

void RSprite16(short x, short y, const unsigned short *sprite, void *vm_addr, short mode, short angle)
{
	unsigned short *scratch, *offset;
	short i, j;

	if(angle == 0)
	{
	  Sprite16(x, y, 16, sprite, vm_addr, mode);
	  return;
	}	
	
	scratch = (unsigned short*)malloc(sizeof(short) * 16);
	memset(scratch, 0, sizeof(short) * 16);
	offset = scratch;
	
	if(angle == 90)
	{
		for(i = 15; i >= 0; i--, offset++)
	  	for(j = 15; j >= 0; j--)
	    	*offset |= ((sprite[j] >> i) & 1) << (15 - j);
	} else if(angle == 270)
	{
		for(i = 15; i >= 0; i--, offset++)
	  	for(j = 15; j >= 0; j--)
	    	*offset |= ((sprite[j] >> (15 - i) & 1)) << (15 - j);
	} else if(angle == 180)
	{
		for(i = 0; i < 16; i++)
		{
		  unsigned short temp = 0;
		  for(j = 0; j < 16; j++)
		    temp |= ((sprite[i] >> j) & 1) << (15 - j);
			scratch[15 - i] = temp;
		}
	}
	    
	Sprite16(x, y, 16, scratch, vm_addr, mode);
		  
	free(scratch);
	return;
}
